// Based on: http://bildr.org/2012/11/sht15-arduino/

#ifndef sht_h
#define sht_h

#ifndef Arduino_h
#include "Arduino.h"
#endif

#ifndef __MATH_H
#include "math.h"
#endif

// Temperature constants for 5V, Celcius, 14-bit
#define SHT_T_D1 ((float)(-40.1))
#define SHT_T_D2 ((float)(0.01))

// Relative humidity constants for 12-bit
#define SHT_RH_C1 ((float)(-2.0468))
#define SHT_RH_C2 ((float)(0.0367))
#define SHT_RH_C3 ((float)(-1.5955E-6))

// Humidity compensation
#define SHT_RH_T1 ((float)(0.01))
#define SHT_RH_T2 ((float)(0.00008))

// Dew point
#define SHT_DEW_TN ((float)(243.12))
#define SHT_DEW_M  ((float)(17.62))


// SHT15 commands
#define SHT_CMD_READ_TEMP  3    // 00000011
#define SHT_CMD_READ_RH    5    // 00000101
#define SHT_CMD_READ_REG   7    // 00000111

const byte sht_crc_table[256] = { 0, 49, 98, 83, 196, 245, 166, 151, 185, 136, 219, 234, 125, 76, 31, 46, 67, 114, 33, 16, 135, 182, 229, 212, 250, 203, 152, 169, 62, 15, 92, 109, 134, 183, 228, 213, 66, 115, 32, 17, 63, 14, 93,108, 251, 202, 153, 168, 197, 244, 167, 150, 1, 48, 99, 82, 124, 77, 30, 47, 184, 137, 218, 235, 61, 12, 95, 110, 249, 200, 155, 170, 132, 181, 230, 215, 64, 113, 34, 19, 126, 79, 28, 45, 186, 139, 216, 233, 199, 246, 165, 148, 3, 50, 97, 80, 187, 138, 217, 232, 127, 78, 29, 44, 2, 51, 96, 81, 198, 247, 164, 149, 248, 201, 154, 171, 60, 13, 94, 111, 65, 112, 35, 18, 133, 180, 231, 214, 122, 75, 24, 41, 190, 143, 220, 237, 195, 242, 161, 144, 7, 54, 101, 84, 57, 8, 91, 106, 253, 204, 159, 174, 128, 177, 226, 211, 68, 117, 38, 23, 252, 205, 158, 175, 56, 9, 90, 107, 69, 116, 39, 22, 129, 176, 227, 210, 191, 142, 221, 236, 123, 74, 25, 40, 6, 55, 100, 85, 194, 243, 160, 145, 71, 118, 37, 20, 131, 178, 225, 208, 254, 207, 156, 173, 58, 11, 88, 105, 4, 53, 102, 87, 192, 241, 162, 147, 189, 140, 223, 238, 121, 72, 27, 42, 193, 240, 163, 146, 5, 52, 103, 86, 120, 73, 26, 43, 188, 141, 222, 239, 130, 179, 224, 209, 70, 119, 36, 21, 59, 10, 89, 104, 255, 206, 157, 172 };


class sht15_sensor {

private:
	int clock_pin;      // pin used for clock
	int data_pin;      // pin used for data
	
	char error_str[64];   // error message for anyone curious

public:

	sht15_sensor(int c, int d) {
		clock_pin = c;
		data_pin = d;
		clear_error();
	}

	int set_error(char *s) {
		strncpy(error_str, s, 63);
		return 0;
	}


	void clear_error(void) {
		*error_str = 0;
	}


	char *get_error(void) {
		return error_str;
	}


	int send_command(int command) {
		clear_error();

		// send a command to the SHTx sensor
		// transmission start
		pinMode(data_pin, OUTPUT);
		pinMode(clock_pin, OUTPUT);
		digitalWrite(data_pin, HIGH);
		digitalWrite(clock_pin, HIGH);
		digitalWrite(data_pin, LOW);
		digitalWrite(clock_pin, LOW);
		digitalWrite(clock_pin, HIGH);
		digitalWrite(data_pin, HIGH);
		digitalWrite(clock_pin, LOW);

		// shift out the command (the 3 MSB are address and must be 000, the last 5 bits are the command)
		shiftOut(data_pin, clock_pin, MSBFIRST, command);

		// verify we get the right ACK
		digitalWrite(clock_pin, HIGH);
		pinMode(data_pin, INPUT);

		if (digitalRead(data_pin)) return set_error("SHT ACK#0");
		digitalWrite(clock_pin, LOW);
		if (!digitalRead(data_pin)) return set_error("SHT ACK#1");

		return 1;
	}


	int wait_for_result(void) {
		clear_error();

		// wait for the SHTx answer
		pinMode(data_pin, INPUT);

		int ack; //acknowledgement

		//need to wait up to 2 seconds for the value
		for (int i = 0; i < 1000; i++) {
			delay(2);
			ack = digitalRead(data_pin);
			if (ack == LOW) break;
		}

		if (ack == HIGH) return set_error("SHT ACK#2");
		
		return 1;
	}


	int read_data(int *data, int *crc) {
		clear_error();

		// get the MSB (most significant bits)
		pinMode(data_pin, INPUT);
		pinMode(clock_pin, OUTPUT);
		byte MSB = shiftIn(data_pin, clock_pin, MSBFIRST);

		// send the required ACK
		pinMode(data_pin, OUTPUT);
		digitalWrite(data_pin, HIGH);
		digitalWrite(data_pin, LOW);
		digitalWrite(clock_pin, HIGH);
		digitalWrite(clock_pin, LOW);

		// get the LSB (less significant bits)
		pinMode(data_pin, INPUT);
		byte LSB = shiftIn(data_pin, clock_pin, MSBFIRST);
		*data = ((MSB << 8) | LSB); //combine bits

		// send the required ACK
		pinMode(data_pin, OUTPUT);
		digitalWrite(data_pin, HIGH);
		digitalWrite(data_pin, LOW);
		digitalWrite(clock_pin, HIGH);
		digitalWrite(clock_pin, LOW);

		// get CRC
		pinMode(data_pin, INPUT);
		*crc = shiftIn(data_pin, clock_pin, MSBFIRST);

		return 1;
	}


        int reverse_byte(int b) {
                int r = 0;
                
                for (int i = 0; i < 8; i++) {
                        r = (r << 1) | (b & 1);
                        b = b >> 1;
                }
                
                return r;
        }


	int verify_crc(int command, int data, int crc) {
		byte c;

		c = 0;
		c ^= command & 0xFF;
                c = sht_crc_table[c];
		c ^= (data >> 8) & 0xFF;  // MSB
                c = sht_crc_table[c];
		c ^= data & 0xFF;         // LSB
                c = sht_crc_table[c];
                c = reverse_byte(c);

                if (c != crc) return set_error("SHT CRC");
                return 1;
	}


	int poll_sensor(int command, int *data) {
		int val;
		int crc;

		clear_error();
		if (!send_command(command)) return 0;
		if (!wait_for_result()) return 0;
		if (!read_data(&val, &crc)) return 0;
                if (!verify_crc(command, val, crc)) return 0;
		*data = val;
		return 1;
	}

	

	int read_temperature(float *so_t, float *temp_c) {
		int val;

		if (!poll_sensor(SHT_CMD_READ_TEMP, &val)) return 0;

                if (so_t)   *so_t = val;
                if (temp_c) *temp_c = SHT_T_D1 + SHT_T_D2 * (float)val;
                return 1;
	}

	
	int read_humidity(float *so_rh, float *rh_linear) {
		int val;
  
		if (!poll_sensor(SHT_CMD_READ_RH, &val)) return 0;

                if (so_rh)     *so_rh = val;
		if (rh_linear) *rh_linear = SHT_RH_C1 + SHT_RH_C2 * (float)val + SHT_RH_C3 * (float)val * (float)val;
                return 1;
	}


        float rh_compensate(float temp_c, float so_rh, float rh_linear) {
                 return (temp_c - 25) * (SHT_RH_T1 + SHT_RH_T2 * so_rh) + rh_linear;
        }
        
        
        float dew_point(float temp_c, float rh_true) {
                  float a = log(rh_true / 100) + (SHT_DEW_M * temp_c) / (SHT_DEW_TN + temp_c);
                  float b = SHT_DEW_M - log(rh_true / 100) - (SHT_DEW_M * temp_c) / (SHT_DEW_TN + temp_c);
                  return SHT_DEW_TN * a / b;
        }




}; // class

#endif
